package complicit

import scala.collection.JavaConverters._
import scala.util.matching.Regex

import soot.{Unit => SootUnit, util => _, _}
import soot.toolkits.graph._
import soot.jimple._
import soot.util.cfgcmd.CFGToDotGraph
import soot.util.dot._
import soot.util.{JasminOutputStream}
import soot.options.Options

import java.io.{PrintWriter, OutputStreamWriter, FileOutputStream}

class GlobalTransform extends SceneTransformer {

  var pcField: SootField = null
  var upgradeMethods: Map[Type, SootMethod] = Map()
  var getTaintMethods: Map[Type, SootMethod] = Map()
  var declassifyMethods: Map[Type, SootMethod] = Map()

  var outMethod: SootMethod = null
  var helperClass: SootClass = null

  protected override def internalTransform(phase: String,
    options: java.util.Map[String, String]) = {
    val mainClass = try {
      Scene.v.getMainClass()
    } catch {
      case e: RuntimeException =>
        // no main class set, typically the case in APKs:
        Scene.v.getApplicationClasses.asScala.head
    }
    val pcFieldName = findFreshFieldName(mainClass)
    pcField = new SootField(pcFieldName, IntType.v, Modifier.PUBLIC | Modifier.STATIC)
    mainClass.addField(pcField)
    helperClass = Scene.v.getSootClass("com.example.daniel.tainted.TaintDroidSinkUtil")
    addOutMethod(mainClass)
    findDeclassifyMethods()
    // todo: ensure freshness of method name
    Util.debug(s"Added pc field to ${mainClass.getName}: ${pcField.getName}" +
      s"[${pcField.getSignature}]")
    findTaintMethods()
    Util.debug(s"Finished initial transformation pass")
  }

  def addOutMethod(mc: SootClass): Unit = {
    outMethod = helperClass.getMethod("testOutput", Seq[Type](IntType.v).asJava,
      VoidType.v)
  }

  def findDeclassifyMethods(): Unit = {
    val types: Set[Type] = Set(IntType.v, FloatType.v, DoubleType.v)
    types foreach { t =>
      declassifyMethods += ((t, helperClass.getMethod("declassify", Seq(t).asJava, t)))
    }
  }

  // not used anymore:
  def addOutMethodManual(mainClass: SootClass): Unit = {
    val methName = s"outMethod${Math.abs((new java.util.Random()).nextInt())}"
    outMethod = new SootMethod(methName, Seq(pcField.getType).asJava, VoidType.v,
      Modifier.PUBLIC | Modifier.STATIC)
    val body = Jimple.v.newBody(outMethod)
    mainClass.addMethod(outMethod)
    outMethod.setActiveBody(body)
    val lg = new soot.javaToJimple.LocalGenerator(body)
    val paramLoc = lg.generateLocal(pcField.getType)
    body.getUnits.addFirst(
      Jimple.v.newIdentityStmt(paramLoc, Jimple.v.newParameterRef(pcField.getType, 0)))
    val smsManagerClass = Scene.v.loadClassAndSupport("android.telephony.SmsManager")
    val smsLoc = lg.generateLocal(smsManagerClass.getType)
    body.getUnits.addLast(Jimple.v.newAssignStmt(smsLoc,
      Jimple.v.newNewExpr(smsManagerClass.getType)))
    val initMeth = smsManagerClass.getMethod("<init>", Seq().asJava)
    body.getUnits.addLast(Jimple.v.newInvokeStmt(
      Jimple.v.newSpecialInvokeExpr(smsLoc, initMeth.makeRef)))
    val pendingIntentClass = Scene.v.loadClassAndSupport("android.app.PendingIntent")
    val strType = RefType.v("java.lang.String")
    val sendTextMeth = smsManagerClass.getMethod("sendTextMessage",
      Seq[Type](strType, strType, strType,
        pendingIntentClass.getType, pendingIntentClass.getType).asJava)
    val integerClass = Scene.v.loadClassAndSupport("java.lang.Integer")
    val toStringMeth = integerClass.getMethod("toString", Seq[Type](IntType.v).asJava)
    val strLoc = lg.generateLocal(RefType.v("java.lang.String"))
    body.getUnits.addLast(Jimple.v.newAssignStmt(strLoc,
      Jimple.v.newStaticInvokeExpr(toStringMeth.makeRef, paramLoc)))
    body.getUnits.addLast(Jimple.v.newInvokeStmt(
      Jimple.v.newVirtualInvokeExpr(smsLoc, sendTextMeth.makeRef,
        StringConstant.v("+4642424242"), NullConstant.v, strLoc,
        NullConstant.v, NullConstant.v)))
    body.getUnits.addLast(Jimple.v.newReturnVoidStmt)
  }

  def findTaintMethods(): Unit = {
    val taintClass = Scene.v.loadClassAndSupport("dalvik.system.Taint")
    val types = Seq(IntType.v, FloatType.v, BooleanType.v, RefType.v("java.lang.String"),
      DoubleType.v, CharType.v)
    types foreach { t =>
      val typeName = if (t == RefType.v("java.lang.String")) "String"
                     else t.toString.capitalize
      val upgMeth = taintClass.getMethod(s"addTaint${typeName}", Seq(t, IntType.v).asJava)
      upgradeMethods += ((t, upgMeth))
      val getMeth = taintClass.getMethod(s"getTaint${typeName}", Seq(t).asJava, IntType.v)
      getTaintMethods += ((t, getMeth))
    }
    Util.debug(s"Found upgrade methods: $upgradeMethods")
    Util.debug(s"Found getTaint methods: $getTaintMethods")
  }

  def isApplicationClass(cl: SootClass): Boolean = {
    !cl.toString.startsWith("android.support") &&
    !Main.options.ignoreClasses.contains(cl.getName)
  }

  def isApplicationMethod(meth: SootMethod): Boolean = {
    isApplicationClass(meth.getDeclaringClass) &&
      (Scene.v.getApplicationClasses.asScala exists (_.getMethods.asScala contains meth))
  }

  def findFreshFieldName(c: SootClass): String = {
    var idx = 0
    val pref = "IfcPC"
    while (c.getFields.asScala.exists(_ == s"${pref}${idx}")) {
      idx += 1
    }
    pref + idx
  }

  def isHelperMethod(sm: SootMethod): Boolean =
    sm.equals(outMethod) || isDeclassifyMethod(sm)

  def isDeclassifyMethod(m: SootMethod): Boolean =
    (declassifyMethods.values exists (_ == m)) ||
      m.getName.contains("declassify")


}

object GlobalTransform {
  private val instance: GlobalTransform = new GlobalTransform()
  def v: GlobalTransform = instance

  def inputTypes: Seq[Type] = Seq(IntType.v, BooleanType.v,
    ShortType.v, LongType.v, FloatType.v, ByteType.v,
    DoubleType.v, RefType.v("java.lang.String"), RefType.v("java.lang.StringBuffer"),
    RefType.v("java.lang.Float"), RefType.v("java.lang.Double"), RefType.v("java.lang.Integer"),
    CharType.v)

  def isInputType(t: Type): Boolean =
    t match {
      case at: ArrayType => isInputType(at.getElementType)
      case t if inputTypes.contains(t) => true
      case _ => false
    }
}
