package complicit

import soot.toolkits.graph._
import soot.Body
import soot.util.cfgcmd.CFGToDotGraph
import java.io.{File, FileOutputStream, PrintWriter}
import sys.process._
import scala.util.matching._
import scalatags.Text.all._
import scalatags.Text._
import scalatags.Text.tags2.title
import soot.util.dot._
import collection.mutable.{MultiMap, Map, Buffer, LinkedHashMap, Set, LinkedHashSet}
import soot.{Unit => SootUnit, util => _, _}
import soot.jimple._
import scalacss.Defaults._
import scalacss.ScalatagsCss._
import java.util.Calendar;
import collection.JavaConverters._

class OverviewItem(val meth: SootMethod,val phase: String, val graph: DotGraph,
  val unitCnt: Integer) {
  def render: TypedTag[String] = {
    Util.debug(s"Rendering $meth.getName")
    val file = Util.renderDotGraph(s"$meth - $phase", graph)
    td("valign".attr := "top", img(src := file))
  }

}

class Overview() {

  var overview: Seq[TypedTag[String]] = Seq();
  var lastMeth: SootMethod = null

  val items: MultiMap[SootMethod, OverviewItem] =
    new LinkedHashMap[SootMethod, Set[OverviewItem]] with
        MultiMap[SootMethod, OverviewItem]

  def logMethod(meth: SootMethod, phase: String, body: Body): Unit = {
    if (
      (meth.getDeclaringClass.getName.contains("CaloriesNotifier") &&
        meth.getName.contains("onStep"))) {
      Util.warn(s"Not drawing giant method ${meth.getName}")
    } else {
      items.addBinding(meth, new OverviewItem(meth, phase,
        Util.makeDotGraph(new BriefUnitGraph(body)),
        body.getUnits.iterator.asScala filter (!_.isInstanceOf[NopStmt]) length))
    }
  }

  def createOverview(): File = {
    val tempFile: File = File.createTempFile("_taintifcOverview", ".html")
    val methodEntries = items flatMap { case (meth, elems) =>
      val elemsSorted = elems.toSeq sortWith (_.phase < _.phase)
      val elemEntries: Seq[TypedTag[String]] = elemsSorted map (_.render)
      val ratio = elemsSorted match {
        case Seq(el1, el2) =>
          Seq(p(s"Overhead: \\( \\frac{${el2.unitCnt}}{${el1.unitCnt}} = " ++
            s"${el2.unitCnt / el1.unitCnt.toDouble} \\)"))
        case _ => Seq()
      }
      Seq(h2(meth.toString), a(name := meth.toString,
        table(tr(elemEntries.toSeq)))) ++ ratio
    }

    /*
    val tocBox = div(myStyle.tocBox, h2("Methods"),
      ul(items.toSeq map {
        case (meth, _) => li(
          a(href := s"#$meth", meth.toString))
      }))
     */
    val cwd = new File(".").getCanonicalPath
    val mathJaxInclude = if (new File(cwd ++ "/MathJax/MathJax.js").exists)
      s"$cwd/MathJax/MathJax.js?config=TeX-AMS-MML_HTMLorMML" else
      "https://cdn.mathjax.org/mathjax/latest/MathJax.js?config=TeX-AMS-MML_HTMLorMML"
    val tags = html(myStyle.render[TypedTag[String]],
      head(title("Complicit"),
        script("type".attr := "text/javascript", src := mathJaxInclude)),
      body(myStyle.bodyStyle, // tocBox,
        h1(s"Transformation Overview (generated ${Calendar.getInstance.getTime()})"),
        methodEntries.toSeq))
    new PrintWriter(tempFile) { write(tags.toString); close }
    Seq("xdg-open", tempFile.getAbsolutePath).!!
    tempFile
  }
  object myStyle extends StyleSheet.Inline {
    import dsl._
    val bodyStyle = style(position.relative)
    val cfgBox = style(verticalAlign.top)
    val tocBox = style(position.fixed, top :=! "20px", backgroundColor :=! "#eeeeee",
      minWidth :=! "300px", maxWidth :=! "450px",
      padding :=! "20px", float.right, border.dashed,
      borderWidth :=! "1", right :=! "20px")
  }
}

object Overview {
  def apply() = new Overview()
}
